-- $Source: /user/cp_source/ObjectAda/rts/common/tasking/system.rts.tasking_types_pkg.spc,v $
-- $Revision: 1.14.1.2 $ $Date: 2001/11/27 20:45:36 $ $Author: shancher $
--        Copyright (C) 1995 by Intermetrics, Inc.

with Ada.Exceptions;
with System.RTS.Exception_Types;
with System.RTS.Interface_Types;
with System.RTS.Protected_Types_Pkg;
with System.RTS.TGT.Frames;
with System.RTS.TGT.Kernel;
with System.RTS.TGT.Kernel.Threads;
with System.RTS.Timer_Class;
with Ada.Finalization;

package System.RTS.Tasking_Types_Pkg is

    pragma Suppress( Elaboration_Check );

--* OVERVIEW
-- This package contains the Task Control Block (TCB) and component types.
-- The activation record and master record are also defined here because
-- of the interdependencies between TCBs and there data structures.
--
-- A TCB contains master components: (1) the enclosing master and (2) the
-- "innermost master" that the task is executing.  A master record
-- contains a pointer to the TCB that is executing it.

-- A TCB contains a pointer to the activation record for the
-- group activation that is currently in progress (if any).
-- This component is need so that the a newly-activated task can
-- notify its master.  The activation record contains a list of
-- tasks to be activated.

    package PRT renames Protected_Types_Pkg;
    package INT renames Interface_Types;

    subtype ECD_Access is PRT.ECD_Access;
    subtype Task_Entry_Index is INT.Task_Entry_Index;
    subtype Timer_Class_Container is Timer_Class.Timer_Class_Container;

    type TCB_Type( Num_Entries : Task_Entry_Index);
    type TCB_Access is access all TCB_Type;
    type Boolean_Ptr is access all Boolean;

    -- Generation number is kept for dangling pointer check.
    -- Generated code stores values of TCB_Handle type.  When a task
    -- terminates, it's storage is deallocated.  To detect the use of
    -- a handle for a terminated task, the generation numbers in the
    -- encloser and TCB proper are compared.

    type TCB_Gen_Num is new Natural;

    -- First two values of TCB_Gen_Num used for null and environment task.
    NOT_IN_USE             : constant TCB_Gen_Num  := TCB_Gen_Num'First;
    TCB_Gen_Count          : TCB_Gen_Num           := TCB_Gen_Num'First + 1;
                                -- Incremented before use

    type Task_Obj_Header is record
        TCB     : TCB_Access;
        Gen_Num : TCB_Gen_Num := NOT_IN_USE;
    end record;

    type Task_Obj_Access is access all Task_Obj_Header;
    -- Note: We are depending on static initialization of the following
    -- special object headers
    Null_Task_Obj_Header : constant
        Task_Obj_Header := Task_Obj_Header'( TCB => null,
                                             Gen_Num => TCB_Gen_Num'First);

    -- Task lists are used for activations, dependencies and free
    -- TCBs.  Links are defined in TCB.  Only one pointer is needed
    -- in the head since none of these lists must be sorted.

    type Task_List_Head is record
       Count : Natural := 0;
       First : TCB_Access;
    end record;

    subtype Task_List_Type is Task_List_Head;

      -- Note that Runnable may be the status even after the task has
      -- been aborted or has completed, because the task is
      -- performing cleanup activities.

    type Task_Status is (
        Unactivated,      -- Created & elaborated; on activation list.
        Runnable,         -- Not suspended in kernel.  Distinction between
                          -- runnable and running is made in kernel.
        --
        -- The remaining states (except Unknown) are a breakdown of the
        -- Suspended State in the kernel.
        --
        Task_Wait,              -- Waiting for dependents to terminate.
        Terminated,             -- Terminated; all dependents terminated;
                                -- no longer running stack space released.
        Delayed,                -- Waiting for delay to expire; on delay queue.
        Activator_Waiting,      -- Waiting for activation of subtasks;
        Caller_Waiting,         -- Normal entry call; on entry queue for
                                -- either protected object or task.
        Caller_Delayed,         -- Timed entry call; on delay and entry queue;
        Acceptor_Waiting,       -- At ACCEPT, awaiting a caller; entries open.
        Acceptor_Delayed,       -- SELECTIVE WAIT with delay; on delay
                                -- queue; entries open.
        Acceptor_Terminable     -- SELECTIVE WAIT with terminate; entries open.
    );

    -- Completion Status

    type Completion_Status_Enum is (
        Not_Yet_Completed,       -- Initial value
        Completed_Normally,      -- Normal completion of task body
        Completed_In_Select,     -- Completion due to selecting
                                 -- terminate alternative.
        Never_Activated,         -- Task terminated without activating.
        Unhandled_Exception,     -- Completion due to unhandled exception.
        Aborted                  -- Completed due to abort of task
                                 -- or ancestor.
    );

    subtype Alternative_Index is INT.Accept_Alt_Index;
      -- index of accept alternative in a select statement

    type Entry_For_Alt is array( Alternative_Index range <> )
      of Task_Entry_Index;
      -- Array that indicates which entry is accepted by each accept
      -- alternative; zero means accept alternative is closed

    type Select_Record( Last_Alt : Alternative_Index );
    type Select_Access is access all Select_Record;

    type Select_Record( Last_Alt : Alternative_Index ) is record
        Has_Terminate   : Boolean;
        Has_Else        : Boolean;
        Delay_Alt_Index : Alternative_Index'Base;
        Rel_Delay_Value : Duration;
        Time_Type       : INT.Time_Type_Enum;
        Abs_Delay_Value : aliased Timer_Class_Container;
        -- Indicates accept alternative for best caller
        Alt_Index_Chosen : Alternative_Index'Base;
          pragma Atomic(Alt_Index_Chosen);  -- Used by callers and acceptor
        Some_Alternative_Open : Boolean; -- accept or delay
        Entry_List      : Entry_For_Alt( 0 .. Last_Alt );
        -- Entry_List is last field so we do not allocate an extra
        -- word for a pointer to the dynamic part.
    end record;

    type Pr_TCB_Type;

    type Pr_TCB_Type_Access is access all Pr_TCB_Type;
      -- needed as target of type cast from PO_Access.
    for Pr_TCB_Type_Access'Storage_Size use 0;
      -- Suppress registration of collection for cleanup; no allocators used

    type Attribute_Handle is new Address;

    type Attribute_ID is new Positive;

    Last_Attribute_ID_Preallocated : constant := 4;
      -- Number of attribute pointers always allocated in TCB

    type Attribute_Index_Preallocated is array(Attribute_ID
        range 1..Last_Attribute_ID_Preallocated) of
        Attribute_Handle;

    type Attribute_Index_Extension is array(Attribute_ID range <>) of
        Attribute_Handle;

    type Attribute_Index_Extension_Access is access Attribute_Index_Extension;

    type Attribute_Cleanup_Op is access protected procedure(
        Handle : Attribute_Handle);

    function Get_Active_Priority( TCB : TCB_Access )
      return System.Any_Priority'Base;
    pragma Inline( Get_Active_Priority );
    -- ??TBD: Pragmas Inline are commented out, to avoid having
    -- to recompile bodies too much.
    --* OVERVIEW
    -- Function  to read Active_Priority attribute.

    procedure Restore_Active_Priority (
      TCB : TCB_Access; Old_Priority : System.Any_Priority'Base);
    --* OVERVIEW
    -- Restore priority to old value at end of rendezvous

    -- For certain atomicly accessible items in the Pr_TCB, read-only
    -- access is provided via the TCB.  These are items that have
    -- individual access functions with no checks.
    -- NOTE: This is an optimization in most cases, however it is
    --   *required* by Call_Entry_Body to avoid deadlock when
    --   checking the Aborting_Level.

    type Internal_View_Of_Pr_TCB_Items is record
        -- 'Callable

        Callable        : Boolean := True;
        pragma Atomic( Callable );

        -- Must include priority because self and other thread
        -- could be setting priority at same time, though this
        -- is highly unlikely.
        -- RTK doesn't keep this

        Base_Priority   : System.Any_Priority;
        pragma Atomic( Base_Priority );

        -- matches System.RTS.TGT.ERTK base priority disregarding locks

        -- Next two fields are protected because multiple triggers
        -- are possible with nested asynchronous selects.

        Aborting_ECD    : ECD_Access;
        pragma Atomic( Aborting_ECD );

        Aborting_Level  : Natural;
        pragma Atomic( Aborting_Level );
    end record;

    -- access type used to give read-only access to above record
    type Pr_TCB_External_View_Ptr is access
      constant Internal_View_Of_Pr_TCB_Items;

    function Is_Callable( TCB : TCB_Access ) return Boolean;
    pragma Inline( Is_Callable );
    --* OVERVIEW
    -- Function  to read Callable attribute.

    function Get_Base_Priority( TCB : TCB_Access ) return System.Any_Priority;
    pragma Inline( Get_Base_Priority );
    --* OVERVIEW
    -- Function  to read Base_Priority attribute.

    function Get_Aborting_Level( TCB : TCB_Access ) return Natural;
    pragma Inline( Get_Aborting_Level );
    --* OVERVIEW
    -- Returns the outermost level to which the task is
    -- aborting.

    function Get_Aborting_ECD( TCB : TCB_Access ) return ECD_Access;
    pragma Inline( Get_Aborting_ECD );
    --* OVERVIEW
    -- Returns the ECD of the outermost level to which the task is
    -- aborting.  Null if an abort is not in progress.

    protected type Pr_TCB_Type (
        Num_Entries : Task_Entry_Index
    ) is
    --* OVERVIEW
    -- This type is the "protected" part of the TCB.
    -- Components are included here if they may be
    -- accessed by more than one task.  Also, protected
    -- operations provide the basis for rendezvous.

        pragma Interrupt_Priority( TGT.Kernel.RTS_Ceiling_Priority );

        procedure Default_Reinitialize(Enclosing_TCB : TCB_Access);
        --* OVERVIEW
        -- Reinitializes to same values as is done for a new object,
        -- allowing re-use.

        entry Task_Entry(1..Num_Entries)(
            Params : PRT.Param_Block_Ptr
        );
        --* OVERVIEW
        -- A task entry call becomes a call on a member
        -- of this family.

        procedure Simple_Accept(
            Select_Info : Select_Access
        );
        --* OVERVIEW
        -- Called for an accept statement not nested in selective
        -- accept.

        procedure Trivial_Accept(
            Which_Entry : Task_Entry_Index
        );
        --* OVERVIEW:
        --   Set Trivial_Entry_Index non-zero; after call
        --   must wait on Caller_Ready suspension object

        procedure Selective_Accept(
            Select_Info : Select_Access
        );
        --* OVERVIEW
        -- A selective accept becomes a call on this entry.
        -- If the selective accept has an open delay alternative,
        -- then a timed entry call on this entry is used.
        -- If the selective accept has an open alternative a simple
        -- call on this entry is used.
        -- The call is followed by a "case" statement to execute
        -- the appropriate accept alternative.
        --* EFFECTS
        -- If there is a caller, then the highest priority caller is chosen.
        -- In the event of a tie, the one calling the entry appearing
        -- in an earlier alternative is chosen.
        -- Actual parameters are obtained from ECD.
        -- Notify master that this task is terminable if
        -- no callers and Terminable = True.

        function Is_Callable return Boolean;
        --* OVERVIEW
        -- Function to read Callable attribute.
        -- OBSOLESCENT (use un-protected function above)

        procedure Set_Uncallable;
        --* OVERVIEW
        -- Sets Callable attribute to False;

        function Get_Base_Priority return System.Any_Priority;
        --* OVERVIEW
        -- Function to read Base_Priority attribute.
        -- OBSOLESCENT (use un-protected function above)

        function Get_Active_Priority return System.Any_Priority'Base;
        --* OVERVIEW
        -- Function to read Base_Priority attribute.
        -- OBSOLESCENT (use un-protected function above)

        procedure Set_Base_Priority(Priority : System.Any_Priority);
        --* OVERVIEW
        -- Sets Base_Priority attribute.

        procedure Inherit_Priority(Priority : System.Any_Priority'Base);
        --* OVERVIEW
        -- Sets Active_Priority to the max of inherited priorities.
        -- Note: Excepting priority inheritance during protected
        -- actions, we may only have a single source of inheritance.

        procedure Disinherit_Priority(Priority : System.Any_Priority'Base);
        --* OVERVIEW
        -- Sets Active_Priority to Base_Priority.

        function Get_Aborting_Level return Natural;
        --* OVERVIEW
        -- Returns the outermost level to which the task is
        -- aborting.
        -- OBSOLESCENT (use un-protected function above)

        procedure Reset_Aborting_Level;
        --* OVERVIEW
        -- Resets the aborting level on a successful abort.

        function Get_Aborting_ECD return ECD_Access;
        --* OVERVIEW
        -- Returns the ECD of the outermost level to which the task is
        -- aborting.  Null if an abort is not in progress.
        -- OBSOLESCENT (use un-protected function above)

        procedure Abort_Task;
        --* OVERVIEW
        -- Procedure to abort a task.

        procedure Abort_To_Level(
            ECD : ECD_Access
        );
        --* OVERVIEW
        -- Procedure to abort an abortable part.  If level in ECD
        -- is lower than current abort in progress, if any, then
        -- this task is sent an abort signal.

        procedure Hold;
        --* OVERVIEW
        --  Called by Ada.Dynamic_Priorities.Hold.

        procedure Continue;
        --* OVERVIEW
        -- Called by Ada.Dynamic_Priorities.Continue.

        function Is_Held return Boolean;
        --* OVERVIEW
        --  Called by Ada.Dynamic_Priorities.Is_Held.

        procedure Recompute_Priorities;
        --* Called when an inherited priority has changed.

        -- operations for Ada.Task_Attributes;
        function Attribute_Reference(ID : Attribute_ID) return Attribute_Handle;

        procedure Attribute_Update(
            Val : in Attribute_Handle;
            ID : Attribute_ID;
            Old_Val : out Attribute_Handle
        );

        procedure Finalize_Attribute(
            ID : Attribute_ID
        );

    private

        function Is_Open(
            Ent : Task_Entry_Index
        ) return Boolean;
        --* OVERVIEW
        -- Called in barrier of Task_Entry.

        -- Protected task state information follows.
        -- Note: explicit default initialization should match the
        -- re-initialization in Default_Reinitialize.

        -- Simple RO shared items.
        -- NOTE: These are directly accessible outside of the protected type
        --   via an access-to-constant pointer.
        --   This is an optimization in most cases, however it is
        --   *required* by Call_Entry_Body to avoid deadlock when
        --   checking the Aborting_Level.
        RO_Items : aliased Internal_View_Of_Pr_TCB_Items;

        -- points to encloser, corresponds to thread ptr. in kernel
        Enclosing_TCB : TCB_Access;

        Current_Select_Info : Select_Access;

        Trivial_Entry_Index : Task_Entry_Index := 0;

        Contributing_Base_Priority : Any_Priority'Base := Any_Priority'First;
        -- base priority that is considered for priority inheritance
        -- same as Base_Priority unless task is held (ATC.Hold).

        Attribute : Attribute_Index_Preallocated :=
            (Others => Attribute_Handle(Null_Address));

        Attribute_Extension : Attribute_Index_Extension_Access;

    end Pr_TCB_Type;

    -- The object below is used to save the type descriptor generated
    -- for Pr_TCB_Type.  It is initialized in RTS_Init and
    -- used to test for special cases involving a Pr_TCB protected object.

    TCB_Pr_Type_Desc_Access : PRT.Pr_Type_Desc_Access;

    type Master_Record;
    type Master_Access is access all Master_Record;

    subtype Frame_ID_Type is TGT.Frames.Frame_ID;

    subtype Dependency_List is Task_List_Type;

    type Cleanup_Record;

    type Cleanup_Ptr is access all Cleanup_Record'Class;
    for Cleanup_Ptr'Storage_Size use 0;
        -- No allocators, and no need to register collection for finalization

    type Cleanup_Record is abstract tagged record
        Previous : Cleanup_Ptr;
    end record;

    procedure Cleanup(Rec : Cleanup_Record) is abstract;

    type Master_Record is record
        Dependents          : Dependency_List;
          -- dependent tasks created by a task declaration
          -- or allocated with allocator type in this master
        Nonterminable_Count : Natural := 1;
          -- used to decide when to accept terminate alternative
          -- and to wait for dependent tasks
        Cleanup_Actions     : Cleanup_Ptr;
          -- actions to cleanup protected objects or objects with user-defined
          -- finalization
        Task_Body_Master    : Boolean := False;
          -- if this master is a task body master, the task must
          -- be marked uncallable upon reaching finalization
        Enclosing_Master    : Master_Access;
          -- dynamically enclosing master
        Enclosing_Task      : TCB_Access;
          -- task executing this master
        Trigger_ECD         : ECD_Access;
          -- the ECD for the trigger in an async select
        Dependents_Complete : TGT.Kernel.Threads.Suspension_Object;
          -- used for awaiting dependent tasks
        Frame_ID            : Frame_ID_Type := TGT.Frames.Null_Frame_ID;
          -- used to identify subprogram master for exception propagation
        Finalize_Errors_To  : Address;
          -- address at which Program_Error should be raised if an
          -- exception is raised during finalization of objects associated
          -- with this master
    end record;         -- Default initializers used for Default_Master

    Default_Master : aliased Master_Record;

    -- The activation record is used to synchronize activity of
    -- activator and the group of tasks that it is currently activating.
    -- A component of this type is kept in the TCB so that activating
    -- tasks can find it for notification.

    type Activation_List;
    type Activation_Access is access all Activation_List;

    type Activation_List is record
        To_Be_Activated        : Task_List_Type;
        Some_Activation_Failed : Boolean := False;
        pragma Atomic(Some_Activation_Failed);
        Activations_Complete   : TGT.Kernel.Threads.Suspension_Object;
    end record;

    type Proc_With_0_Parms is access procedure;  -- ECR 6790

    type TCB_Type( Num_Entries : Task_Entry_Index) is new
            TGT.Kernel.Threads.Thread_Descriptor with record
        -----------------------------------------------------------------
        --- Task Identification ---
        -----------------------------------------------------------------

        Task_Obj        : Task_Obj_Access; -- ptr to task object
        Gen_Num         : TCB_Gen_Num;    -- Copy for validity check
        Static_Link     : INT.Data_Address;
        Code            : INT.Code_Address;
        Storage_Size    : Integer;
        Elab_Bit        : Boolean_Ptr;
        -----------------------------------------------------------------
        --- State Information ---
        -----------------------------------------------------------------
        Status          : Task_Status := Unactivated;
            -- put Status in Pr_TCB_Pkg?
        Completion_Status : Completion_Status_Enum := Not_Yet_Completed;

        Occurrence      : Exception_Types.Minimal_Exception_Occurrence;
        Blocking_ECD    : ECD_Access;
        Activation_Sync : Activation_Access;
          -- assoc. w/ Activator_Waiting
        Async_Level     : Natural := 0;
          -- nesting level of asynchronous selects

        Notify_Activator_of_Abort : Proc_With_0_Parms;
          -- When task has been activated, but has not yet completed
          -- activation, this value is a pointer to a procedure which
          -- should be called to notify the activator that the task
          -- has been aborted  (ECR 6790).

        Active_Priority : System.Any_Priority'Base;
        pragma Atomic( Active_Priority );
          -- NOTE: We moved this out of Pr_TCB so end_rendezvous
          --  didn't need to lock the TCB;
          -- TBD: There might be a race condition here.

        -----------------------------------------------------------------

        --- Scope Information ---
        -----------------------------------------------------------------
        Master          : Master_Access; -- of this task.
        Innermost_Submaster : Master_Access; -- master most
                                                 -- recently entered
        Activator       : TCB_Access;     -- task that act'd this one
        -----------------------------------------------------------------
        --- Rendezvous Info ---
        -----------------------------------------------------------------
        Select_Info     : Select_Access;
          -- really need a discriminant for # accept alternatives

        Current_Call_Info : ECD_Access;
          -- Non-null while caller in rendezvous

        Caller_Ready : TGT.Kernel.Threads.Suspension_Object;
          -- Set true when caller arrives

        -----------------------------------------------------------------
        --- Links ---
        -----------------------------------------------------------------
        Act_Prev        : TCB_Access;     -- activation list
        Act_Next        : TCB_Access;
        Dep_Prev        : TCB_Access;     -- dependency list
        Dep_Next        : TCB_Access;

        -----------------------------------------------------------------
        --- Read-only access to certain protected fields ---
        -----------------------------------------------------------------
        RO_Items_Ptr      : Pr_TCB_External_View_Ptr;

        -----------------------------------------------------------------
        --- Protected part of TCB ---
        -----------------------------------------------------------------
        Pr_TCB  : Pr_TCB_Type(Num_Entries => Num_Entries);
    end record;

    procedure Release_Thread(TCB : access TCB_Type);
        -- This operation is called by Finish_Thread to allow reuse of
        -- a TCB that has been placed in the free list by Free_TCB.

    function Self return TCB_Access;
    --* Returns pointer to TCB of the task that is executing.
    pragma Inline(Self);      -- Allow special treatment of current_thread

    function New_TCB return TCB_Access;
    --* Returns pointer to a new or reused available TCB.

    procedure Free_TCB(TCB : TCB_Access);
    --* Returns TCB to free list marked as still in use.

    procedure Default_Reinitialize_TCB (TCB : TCB_Access);
    --* Needed for init. of envr. TCB

    function Assign_Attribute_ID(
        Cleanup_Op : Attribute_Cleanup_Op
    ) return Attribute_ID;
    --* Registers attribute Cleanup routine in system-wide table and returns
    -- next available ID.

    procedure Deassign_Attribute_ID(
        ID : Attribute_ID
    );
    --* Makes attribute ID reusable.

end System.RTS.Tasking_Types_Pkg;
