-- $Source: /user/cp_source/ObjectAda/rts/common/tasking/system.rts.protected_types_pkg.dat,v $
-- $Revision: 1.6 $ $Date: 1998/04/13 17:42:46 $ $Author: curt $
--        Copyright (C) 1995 by Intermetrics, Inc.

with Ada.Exceptions; 
with System.RTS.TGT.Kernel.Threads;
with System.RTS.TGT.Kernel.ATC;
with System.RTS.TGT.Kernel.Mutexes;
with System.RTS.Timer_Class;
with System.RTS.Protected_Type_Descriptor_Pkg;

pragma Elaborate(System.RTS.TGT.Kernel.Threads);
-- Reason why this pragma is introduced:
--  Whenever a protected object is declared,
--  rts_init_po is called.
--  rts_init_po calls undefer_abort, which eventually
--  calls System.RTS.TGT.Kernel.Threads.lock_kernel, which assumes that the
--  kernel_mutex is initialized.
--  By putting a pragma elaborate on this package, we make sure that
--  this kernel_mutex is initialized whenever a protected type is declared.

package System.RTS.Protected_Types_Pkg is

    pragma Suppress( Elaboration_Check );

--* OVERVIEW
-- This package contains representation types for protected types
-- and objects required by the Patriot Ada RTM. For each protected
-- type declared by the user, an object of type Pr_Type_Desc is
-- allocated and initialized by the compiler. For each protected
-- object declared (or allocated), an object of type PO_Rec is
-- allocated by the compiler as part of the protected object, but
-- initialized by the RTS.  The Pr_Type_Desc structure is shared
-- by all protected objects of the same type.

    subtype Exception_Occurrence_Access is
	Ada.Exceptions.Exception_Occurrence_Access;


    --============ SIMPLE TYPES ============

    type Barrier_Vector is new Integer;

    type Call_Kind_Enum is
      ( Conditional,
      Unconditional,
      Timed,
      Asynchronous );

    type Call_State_Enum is
      ( Not_Started,
      In_Entry_Body,
      Queued_Abortably,
      Queued_Nonabortably,
      Caller_In_Rendezvous,
      Requeue_Another,
      Requeue_Same,
      Requeue_Task_To_Protected,
      Canceled,
      Completed );

    subtype Code_Address is Address;

    subtype Data_Address is Address;

    subtype Entry_Name_Index is 
	  System.RTS.Protected_Type_Descriptor_Pkg.Entry_Name_Index;

    subtype Entry_Index is
	  System.RTS.Protected_Type_Descriptor_Pkg.Entry_Index;

    type Entry_Member_Index is new Integer;

    subtype Master_Handle is Data_Address;
	
    subtype Param_Block_Ptr is Data_Address;

    subtype PT_Kind_Enum is
      System.RTS.Protected_Type_Descriptor_Pkg.PT_Kind_Enum;

    subtype Handler_Index is
      System.RTS.Protected_Type_Descriptor_Pkg.Handler_Index;

    --============ SIMPLE ARRAY TYPES ============

    subtype Code_Array is
      System.RTS.Protected_Type_Descriptor_Pkg.Code_Array;
    subtype Handler_Array is
      System.RTS.Protected_Type_Descriptor_Pkg.Handler_Array;

    --============ RECORD TYPE DECLARATIONS ============

    type ECD_Rec( Kind : Call_Kind_Enum );
    type ECD_Access is access all ECD_Rec;

    type Handler_Desc;
    type Handler_Desc_Access is access all Handler_Desc;

    type Handler_Info;
    type Handler_Info_Access is access all Handler_Info;

    type PO_Rec( PT_Kind     : PT_Kind_Enum;
		 Max_Index   : Entry_Index );
    type PO_Access is access all PO_Rec;

    subtype Pr_Type_Desc is
      System.RTS.Protected_Type_Descriptor_Pkg.Pr_Type_Desc;
 
    type Pr_Type_Desc_Access is access all Pr_Type_Desc;
 
    subtype Handler_List_Type is
      System.RTS.Protected_Type_Descriptor_Pkg.Handler_List_Type;
    subtype Handler_List_Access is
      System.RTS.Protected_Type_Descriptor_Pkg.Handler_List_Access;

    --============ RECORD ARRAY TYPES ============

    type Queue is array( Entry_Index range <> ) of ECD_Access;

    --============ RECORD TYPES ============

    type ECD_Rec( Kind : Call_Kind_Enum ) is record

        -- Input parameters for protected entry calls.

        Called_PO                   : PO_Access;
        Params                      : Param_Block_Ptr;
        Name_Index   		    : Entry_Name_Index;
        Member_Index 		    : Entry_Member_Index;

        -- Input parameters for delay/timed entry calls
        Expiration                  : aliased Timer_Class.Timer_Class_Container;

        -- Used within the RTS

        -- Priority used by queuing algorithm.  Different for
	-- FIFO and Priority Queueing policies
        Priority                    : Any_Priority;
	-- Active priority of caller at time of call.  Updated 
	-- when priority of caller changes except in async case.
        Caller_Priority             : Any_Priority'Base;
	Calling_TCB		    : Data_Address;

        -- Used to restore previous rendezvous level in case if nested
	-- accepts.
	Old_Acceptor_Priority 	    : Any_Priority'Base;
	Old_Caller_Info 	    : ECD_Access;

        -- Links for individual entry queues, Small or Large model.

        Pred                        : ECD_Access;
        Succ                        : ECD_Access;

        -- Link for next queue head, Large model only.

        Prev_Queue                  : ECD_Access;
        Next_Queue                  : ECD_Access;

        -- Count, meaningful if ECD is head of queue

        Count                       : Natural;

        -- Caller waits on this suspension object.  State upon which
        -- caller wants to be woken depends on kind of call.
        -- Queued_Abortably is such a state for Conditional, Timed
        -- and Async and call cancellation.  Reaching Completed
        -- always sets Wake_Up to True.

        Wake_Up                     : TGT.Kernel.Threads.Suspension_Object;
        Wake_Up_If_Queued_Abortably : Boolean := FALSE;
	    -- False for Unconditional; True otherwise

	Call_Is_Abort_Deferred	    : Boolean := FALSE;
	    -- True if abort handling is deferred at point of entry call
	    -- (this info needed for requeue-with-abort per 9.5.4(14))


        Requeue_With_Abort          : Boolean := True;
	Requeue_Called_PO           : PO_Access;
	Requeue_Name_Index          : Entry_Name_Index;
	Requeue_Member_Index        : Entry_Member_Index;
        Requeue_Params              : Param_Block_Ptr;

        -- Output parameters

        State                       : Call_State_Enum := Not_Started;
        Exception_To_Be_Raised      : Boolean         := False;
        Exception_Occurrence        : Exception_Occurrence_Access;

        -- Extra RTS bookkeeping for Asynchronous calls.

        case Kind is
            when Unconditional | Conditional | Timed =>
                null;
            when Asynchronous =>
                Async_Level         : Natural;
                -- The remaining fields in this variant are needed
                -- in case call completes and abortable part must
                -- aborted with an abort signal
                Finalization_Mark   : Master_Handle;
                Saved_State         : TGT.Kernel.ATC.Setjmp_Buffer_Ptr;
                -- Params_Ptr is pointed to by Params above.  Needed as a 
 	        -- safe place to keep pointer-to-pointer to parameter
                -- block for task calls.
                Params_Ptr          : Data_Address;
        end case;

    end record;

    type Handler_Desc is array( 1 .. 7 ) of Integer;

    type Handler_Info is
	record
        PO_Address   : PO_Access;
        Proc_Address : Code_Address;
    end record;

    -- The index of a handler has no use except for the construction
    -- of a simple list, unlike the case of an entry name index.

    type PO_Rec( PT_Kind     : PT_Kind_Enum;
		 Max_Index   : Entry_Index ) is new
      TGT.Kernel.Mutexes.Mutex_Descriptor with record
        Type_Info            : Pr_Type_Desc_Access;
        Static_Link          : Data_Address;
        case PT_Kind is
            when System.RTS.Protected_Type_Descriptor_Pkg.PT_No_Entries =>
                null;
            when System.RTS.Protected_Type_Descriptor_Pkg.PT_Small      =>
                Is_Valid     : Boolean := FALSE;
                Barrier      : Barrier_Vector := 0;
                Small_Queue  : Queue( Entry_Index'First .. Max_Index );
            when System.RTS.Protected_Type_Descriptor_Pkg.PT_Large      =>
                Large_Queue  : ECD_Access;
        end case;
    end record;

end System.RTS.Protected_Types_Pkg;
